#include <fcntl.h>
#include <stdlib.h>
#include <stdio.h>
void *alloc1 (size_t n1, size_t size)
/*****************************************************************************
allocate a 1-d array 
******************************************************************************
Input:
n1		number of items in the array
size		size of each item

Returned:	pointer to the array
*****************************************************************************/
{
	void *p;

	if ((p=malloc(n1*size))==NULL)
		return NULL;
	return p;
}


void *realloc1(void *v, size_t n1, size_t size)
/*****************************************************************************
reallocate a 1-d array 
******************************************************************************
Input:
n1		number of items in the array
size		size of each item
v		pointer to the previously allocated array

Returned:	pointer to the reallocated array
*****************************************************************************/
{
	void *p;

	if ((p=realloc(v,n1*size))==NULL)
		return NULL;
	return p;
}


void free1 (void *p)
/*****************************************************************************
free a 1-d array 
******************************************************************************
Input:
p		pointer to the array
*****************************************************************************/
{
	free(p);
}

void obmemset1 (float *p, float c, int n1)
/*****************************************************************************
initialize a 1-d array 
******************************************************************************
Input:
p		pointer to the array
c		value with which to initialize array
n1		first dimension of array
*****************************************************************************/
{
	int i;

	for (i = 0; i < n1; ++i)
		p[i] = c;
}

void **alloc2 (size_t n1, size_t n2, size_t size)
/*****************************************************************************
allocate a 2-d array 
******************************************************************************
Input:
n1		number of items in 1st dimension of array
n2		number of items in 2nd dimension of array
size		size of each item

Returned:	pointer to the array
*****************************************************************************/
{
	size_t i2;
	void **p;

	if ((p=(void**)malloc(n2*sizeof(void*)))==NULL) 
		return NULL;
	if ((p[0]=(void*)malloc(n2*n1*size))==NULL) {
		free(p);
		return NULL;
	}
	for (i2=0; i2<n2; i2++)
		p[i2] = (char*)p[0]+size*n1*i2;
	return p;
}


void free2 (void **p)
/*****************************************************************************
free a 2-d array 
******************************************************************************
Input:
p		pointer to the array
*****************************************************************************/
{
	free(p[0]);
	free(p);
}


void obmemset2 (float **p, float c, int n1, int n2)
/*****************************************************************************
initialize a 2-d array 
******************************************************************************
Input:
p		pointer to the array
c		value with which to initialize array
n1		first dimension of array
n2		second dimension of array
*****************************************************************************/
{
    int i, j;

    for (j = 0; j < n2; ++j)
        for (i = 0; i < n1; ++i)
            p[j][i] = c;
}

void ***alloc3 (size_t n1, size_t n2, size_t n3, size_t size)
/*****************************************************************************
 allocate a 3-d array 
******************************************************************************
Input:
n1		number of items in 1st dimension of array
n2		number of items in 2nd dimension of array
n3		number of items in 3rd dimension of array
size		size of each item

Returned:	pointer to the array
****************************************************************************/
{
	size_t i3,i2;
	void ***p;

	if ((p=(void***)malloc(n3*sizeof(void**)))==NULL)
		return NULL;
	if ((p[0]=(void**)malloc(n3*n2*sizeof(void*)))==NULL) {
		free(p);
		return NULL;
	}
	if ((p[0][0]=(void*)malloc(n3*n2*n1*size))==NULL) {
		free(p[0]);
		free(p);
		return NULL;
	}
	for (i3=0; i3<n3; i3++) {
		p[i3] = p[0]+n2*i3;
		for (i2=0; i2<n2; i2++)
			p[i3][i2] = (char*)p[0][0]+size*n1*(i2+n2*i3);
	}
	return p;
}


void*** bind3 (void* buf, size_t n1, size_t n2, size_t n3, size_t size)
{
    size_t  i3,i2;
    void*** p;

    if ((p = (void***) malloc (n3 * sizeof (void**))) == NULL) {
	fprintf (stderr, "bind3: Can't Allocate %ld Elements Of %ldB Size!\n", n3, sizeof (void**));
	return NULL;
    }
    if ((p [0] = (void**) malloc (n3 * n2 * sizeof (void*))) == NULL) {
	free (p);
	fprintf (stderr, "bind3: Can't Allocate %ld Elements Of %ldB Size!\n", n3 * n2, sizeof (void*));
	return NULL;
    }
    if ((p [0][0] = buf) == NULL) {
	free (p [0]);
	free (p);
	fprintf (stderr, "bind3: Can't Allocate %ld Elements Of %ldB Size!\n", n3 * n2 * n1, size);
	return NULL;
    }
    for (i3 = 0; i3 < n3; i3++) {
	p [i3] = p [0] + n2 * i3;
	for (i2 = 0; i2 < n2; i2++)
	    p [i3][i2] = (char*) p [0][0] + size * n1 * (i2 + n2 * i3);
    }
    return p;
}

void free3 (void ***p)
/*****************************************************************************
free a 3-d array 
******************************************************************************
Input:
p		pointer to the array
*****************************************************************************/
{
	free(p[0][0]);
	free(p[0]);
	free(p);
}

void obmemset3 (float ***p, float c, int n1, int n2, int n3)
/*****************************************************************************
initialize a 3-d array 
******************************************************************************
Input:
p		pointer to the array
c		value with which to initialize array
n1		first dimension of array
n2		second dimension of array
n3		third dimension of array
*****************************************************************************/
{
    int i, j, k;

	for (k = 0; k < n3; ++k)
		for (j = 0; j < n2; ++j)
			for (i = 0; i < n1; ++i)
				p[k][j][i] = c;
}


void ****alloc4 (size_t n1, size_t n2, size_t n3, size_t n4, size_t size)
/*****************************************************************************
 allocate a 4-d array 
******************************************************************************
Input:
n1		number of items in 1st dimension of array
n2		number of items in 2nd dimension of array
n3		number of items in 3rd dimension of array
n4		number of items in 4th dimension of array
size		size of each item

Returned:	pointer to the array
****************************************************************************/
{
	size_t i4,i3,i2;
	void ****p;

	if ((p=(void****)malloc(n4*sizeof(void***)))==NULL)
		return NULL;
	if ((p[0]=(void***)malloc(n4*n3*sizeof(void**)))==NULL) {
		free(p);
		return NULL;
	}
	if ((p[0][0]=(void**)malloc(n4*n3*n2*sizeof(void*)))==NULL) {
		free(p[0]);
		free(p);
		return NULL;
	}
	if ((p[0][0][0]=(void*)malloc(n4*n3*n2*n1*size))==NULL) {
		free(p[0][0]);
		free(p[0]);
		free(p);
		return NULL;
	}
	for (i4=0; i4<n4; i4++) {
		p[i4] = p[0]+i4*n3;
		for (i3=0; i3<n3; i3++) {
			p[i4][i3] = p[0][0]+n2*(i3+n3*i4);
			for (i2=0; i2<n2; i2++)
				p[i4][i3][i2] = (char*)p[0][0][0]+
						size*n1*(i2+n2*(i3+n3*i4));
		}
	}
	return p;
}


void free4 (void ****p)
/*****************************************************************************
free a 4-d array 
******************************************************************************
Input:
p		pointer to the array
*****************************************************************************/
{
	free(p[0][0][0]);
	free(p[0][0]);
	free(p[0]);
	free(p);
}

