!==============================================================

! These are modules that are used for performing DRLSE (Distance Regularized Level Set Evolution). They were orignally written in MATLAB by Chunming Li (lchunming@gmail.com, i_chunming@hotmail.com ), and converted into fortran by Taylor Dahlke (Stanford University).


! This implementation is based upon the following paper:
! C. Li, C. Xu, C. Gui, M. D. Fox, "Distance Regularized Level Set Evolution and Its Application to Image Segmentation", IEEE Trans. Image Processing, vol. 19 (12), pp.3243-3254, 2010.

!==============================================================
module drlse_mod
use sep
implicit none
real, private    			:: pi, small_number
integer, private    			:: nr, nc, i, j, k
contains
!==============================================================



subroutine drlse_init(nr_in, nc_in)
integer           			:: nr_in, nc_in
nr = nr_in
nc = nc_in
pi = 3.1415926535897932384626433832795028841971
small_number=0.0000000001
end subroutine





subroutine drlse_edge(phi_0, g, lambda, mu, alpha, epsilon, time_step, num_iter, double_well, phi)
real, dimension(:,:) 			:: phi_0, g, phi
real, dimension(nr,nc) 			:: vx, vy, phi_x, phi_y, s, Nx, Ny, curvature, distRegTerm, laplac_mat, dirac_phi, areaTerm, edgeTerm, phi_in, phi_out
integer               			:: num_iter!, time_step
real					:: lambda, mu, alpha, epsilon, time_step
logical					:: double_well

vx=0.0; vy=0.0; phi_x=0.0; phi_y=0.0; s=0.0; Nx=0.0; Ny=0.0; curvature=0.0; distRegTerm=0.0; laplac_mat=0.0; dirac_phi=0.0; areaTerm=0.0; edgeTerm=0.0; phi_in=0.0; phi_out=0.0; phi=0.0; 

phi=phi_0
 call gradient(g, vx, vy)

do k=1,num_iter
	phi_in=phi

	call neumann_bound_cond(phi_in,phi_out)

	call gradient(phi_out, phi_x, phi_y)
	s = sqrt( phi_x**2 + phi_y**2)
	Nx = phi_x/(s + small_number)
	Ny = phi_y/(s + small_number)
	call div(Nx,Ny,curvature)

	!Using single well
	call del2(phi_out,laplac_mat)
	distRegTerm = 4*laplac_mat - curvature

	call dirac(phi_out,epsilon,dirac_phi)
	areaTerm = dirac_phi * g ! Balloon / pressure force
	edgeTerm = dirac_phi * ((vx * Nx) + (vy * Ny)) + dirac_phi * g * curvature
	phi = phi_out + time_step*(mu*distRegTerm + lambda*edgeTerm + alpha*areaTerm)
end do
end subroutine






subroutine reinitialize(phi_0, time_step, num_iter, phi)
real, dimension(:,:) 			:: phi_0, phi
real, dimension(nr,nc) 			:: phi_x, phi_y, s, phi_out, map_of_sign
integer               			:: num_iter, percent
real					:: time_step, p

phi_x=0.0; phi_y=0.0; s=0.0; phi_out=0.0; phi=0.0; 

 call signmap(phi_0,map_of_sign)

do k=1,num_iter

	call sep_write(phi_0)
	phi_out = phi_0
	call neumann_bound_cond(phi_0,phi_out)
	call gradient(phi_out, phi_x, phi_y)
	s = sqrt( phi_x**2 + phi_y**2)


	phi   = phi_out + time_step*( (1-s)*map_of_sign )
	phi_0 = phi

	! Write out the gradient (optional)
 	if (exist_file('gradient')) then
		call sep_write(s,'gradient')
 	end if

	! Calculate progress
 	p = (real(k)/(real(num_iter)+1))*100.0
 	percent = int(p)
 	write(0,*) " ",percent,"% Complete"  
 	write(0,*) " level set test value", phi_0(nr/2, nc/2)

end do
end subroutine






subroutine signmap(phi,map_of_sign)
real, dimension (:,:)  	:: phi,map_of_sign
map_of_sign=0.0;

do j=1,nc 
    do i=1,nr
	if ( phi(i,j) >= 0.0 ) then
		map_of_sign(i,j) =  1.0
	else
		map_of_sign(i,j) = -1.0
	end if
    end do
end do

end subroutine







subroutine gradient(nt_in, nt_dx, nt_dy)
real, dimension(:,:)			:: nt_in, nt_dx, nt_dy
nt_dx=0.0
nt_dy=0.0

do j=1,nc 
    do i=1,nr
	! Middle gradient values
	if(j/=1 .or. j/=nc) then
	nt_dx(i,j) = (nt_in(i,j+1) - nt_in(i,j-1) )/2
	endif
	if(i/=1 .or. i/=nr) then
	nt_dy(i,j) = (nt_in(i+1,j) - nt_in(i-1,j) )/2
	endif
	! Edge dx gradient values
	if(j==1) then
	nt_dx(i,1) = nt_in(i,2) - nt_in(i,1)
	endif
	if(j==nc) then
	nt_dx(i,nc) = nt_in(i,nc) - nt_in(i,nc-1)
	endif
    end do
    ! Edge dy gradient values
    nt_dy(1,j) = nt_in(2,j) - nt_in(1,j)
    nt_dy(nr,j) = nt_in(nr,j) - nt_in(nr-1,j)
end do
end subroutine





subroutine del2(in, out)
real, dimension(:,:) 			:: in, out
real, dimension(nr,nc) 			:: interior, left, right, top, bottom, ul_corner, br_corner 
integer               			:: i, j
real					:: h, ul, ur, bl, br
! Zero out internal arrays
out = 0.0; interior = 0.0; left = 0.0;  right = 0.0;  top = 0.0;  bottom = 0.0;  ul_corner = 0.0;  br_corner = 0.0;
h=2.0

! Interior Points
do j=1,nc
    do i=1,nr
	! Interior Point Calculations
	if( j>1 .and. j<nc .and. i>1 .and. i<nr )then
	interior(i,j) = ((in(i-1,j) + in(i+1,j) + in(i,j-1) + in(i,j+1)) - 4*in(i,j) )/(h**2)
	end if
	! Boundary Conditions for Left and Right edges
	left(i,1) = (-5.0*in(i,2) + 4.0*in(i,3) - in(i,4) + 2.0*in(i,1) + in(i+1,1) + in(i-1,1) - 2.0*in(i,1) )/(h**2)
	right(i,nc) = (-5.0*in(i,nc-1) + 4.0*in(i,nc-2) - in(i,nc-3) + 2.0*in(i,nc) + in(i+1,nc) + in(i-1,nc) - 2.0*in(i,nc) )/(h**2)
    end do
    ! Boundary Conditions for Top and Bottom edges
    top(1,j) = (-5.0*in(2,j) + 4.0*in(3,j) - in(4,j) + 2.0*in(1,j) + in(1,j+1) + in(1,j-1) - 2.0*in(1,j) )/(h**2)
    bottom(nr,j) = (-5.0*in(nr-1,j) + 4.0*in(nr-2,j) - in(nr-3,j) + 2.0*in(nr,j) + in(nr,j+1) + in(nr,j-1) - 2.0*in(nr,j) )/(h**2)
end do
out = interior + left + right + top + bottom 
! Calculate BC for the corners
ul = (-5.0*in(1,2) + 4.0*in(1,3) - in(1,4) + 2.0*in(1,1) - 5.0*in(2,1) + 4.0*in(3,1) - in(4,1) + 2.0*in(1,1))/(h**2)
br = (-5.0*in(nr,nc-1) + 4.0*in(nr,nc-2) - in(nr,nc-3) + 2.0*in(nr,nc) - 5.0*in(nr-1,nc) + 4.0*in(nr-2,nc) - in(nr-3,nc) + 2.0*in(nr,nc))/(h**2)
bl = (-5.0*in(nr,2) + 4.0*in(nr,3) - in(nr,4) + 2.0*in(nr,1) - 5.0*in(nr-1,1) + 4.0*in(nr-2,1) - in(nr-3,1) + 2.0*in(nr,1))/(h**2)
ur = (-5.0*in(1,nc-1) + 4.0*in(1,nc-2) - in(1,nc-3) + 2.0*in(1,nc) - 5.0*in(2,nc) + 4.0*in(3,nc) - in(4,nc) + 2.0*in(1,nc))/(h**2)
! Apply BC for the corners
out(1,1)=ul
out(1,nc)=ur
out(nr,1)=bl
out(nr,nc)=br
end subroutine





subroutine div(nx,ny,out)
real, dimension (:,:)  			:: nx, ny, out
real, dimension (nr,nc)			:: outx, outy, junk
outx=0.0; outy=0.0; junk=0.0; out=0.0;
 call gradient(nx,outx,junk)
 call gradient(ny,junk,outy)
out=outx+outy
end subroutine




subroutine neumann_bound_cond(in,out)
real, dimension(:,:)			:: in, out
out=0.0
out=in
out((/ 1, nr /), (/ 1, nc /)) = out((/ 3, nr-2 /), (/ 3, nc-2 /))
out((/ 1, nr /), 2:(nc-1)) = out((/ 3, nr-2 /), 2:(nc-1))
out(2:(nr-1), (/ 1, nc /)) = out(2:(nr-1), (/ 3, nc-2 /))
end subroutine





subroutine dirac(x,sigma,out)
real					:: sigma
real, dimension (:,:)  			:: x, out
real, dimension (nr,nc)			:: a
a=0; out=0.0;
out=(1/(2*sigma))*(1 + cos(pi*x/sigma))
! Make binary selector matrices
do j=1,nc
    do i=1,nr    
	if(x(i,j) <= sigma .and. x(i,j) >= -sigma) then
		a(i,j) = 1 ! "a" matrix
	endif
    end do
end do
out=out*a
end subroutine






end module
