
module levelset_mod
	use sep
	use calc_mod
	use drlse_mod
	use newton_mod
	use fbi_mod
	implicit none

	integer, private                         	:: nz, nx, iz, ix, nt, rec_nx, sou_nx, i , j, k, smooth_rad
	real, private					:: Vsalt, dt, tstep, vstep, mu, dx, dz, fmaxt, normval
	real, dimension(:,:), pointer, private 		:: phi_in, vel_back_ns, phi_out, vel_out
	real, dimension(:), allocatable, private  	:: alphaarray, betaarray, objarray
	real, dimension(:,:), allocatable, private  	:: vel_full, phi_grad, drlse_term, s
	real, dimension(:,:), allocatable, private  	:: scaling, tomograd, tomograd_ns, junk, phi_update
	real, dimension(:,:,:), allocatable, private  	:: residual, tomodata, bounddata
	contains



!/////////////////////////////////////////////////////////////////////////////////////
	subroutine levelset_init( nz_in, nx_in, Vsalt_in, nt_in, rec_nx_in, sou_nx_in, dx_in, dz_in, dt_in, smooth_rad_in)
		integer		:: nz_in, nx_in, nt_in, rec_nx_in, sou_nx_in, smooth_rad_in
		real		:: Vsalt_in, tstep_in, vstep_in, mu_in, dx_in, dz_in, dt_in
!--------------- Read in variables to module ----------------------
		rec_nx = rec_nx_in 
		sou_nx = sou_nx_in
		smooth_rad = smooth_rad_in
		nt = nt_in
		Vsalt = Vsalt_in
		nz = nz_in
		nx = nx_in
		dx = dx_in
		dz = dz_in
		dt = dt_in
!---------------- Allocate -------------------------------
		call mem_alloc2d(vel_full, nz, nx)
		call mem_alloc2d(phi_grad, nz, nx)
		call mem_alloc2d(s, nz, nx)
		call mem_alloc2d(drlse_term, nz, nx)
		call mem_alloc2d(phi_update, nz, nx)
		call mem_alloc2d(scaling, nz, nx)
		call mem_alloc2d(tomograd, nz, nx)
		call mem_alloc2d(tomograd_ns, nz, nx)
		call mem_alloc2d(junk, nz, nx)
		call mem_alloc3d(residual, nt, rec_nx, sou_nx)
		call mem_alloc3d(tomodata, nt, rec_nx, sou_nx)
		call mem_alloc3d(bounddata, nt, rec_nx, sou_nx)
		fmaxt = minval((/ dx, dz /))
		allocate(alphaarray(1))
		allocate(betaarray(1))
		allocate(objarray(1))
	end subroutine





!/////////////////////////////////////////////////////////////////////////////////////
	subroutine levelset(phi_inA, vel_backA, phi_outA, vel_outA)
		real, dimension(:,:), target	:: phi_inA, vel_backA, phi_outA, vel_outA
		real				:: maxstep, alpha, beta, objfuncval, gamma
		integer				:: radius

		phi_in       => phi_inA
		vel_back_ns  => vel_backA
		phi_out      => phi_outA
		vel_out      => vel_outA

!-----------  Calculate residual based on input Vmodel  ----------------------------
		call calc_velocity(phi_in, vel_back_ns, vel_full, Vsalt)
		call calc_residual(vel_full, residual )
		call dotproduct3(residual, objfuncval, .false., .false.)
		objarray(1) = objfuncval

!-----------  Calculate DRLSE term  ------------------------------------------------
		call calc_drlse_term( phi_in, drlse_term, s )

!-----------  Create the velocity scaling coefficients -----------------------------
		junk = Vsalt
		scaling = junk - vel_back_ns

!-----------  Normalize scaling coefficients ---------------------------------------
		scaling = scaling/maxval(scaling)

!-----------  Calculate phigrad  ---------------------------------------------------
		! Adjoint: residual ==> tomograd --> phigrad
		call phi_grad_operator(phi_grad, residual, tomograd_ns, s, scaling, .true.)

!-----------  Normalize gradients --------------------------------------------------
		normval = 10.0**13.0
		phi_grad = phi_grad/normval
		tomograd_ns = tomograd_ns/normval

!-----------  Find initial step values ---------------------------------------------
		maxstep = fmaxt/maxval(abs(phi_grad))
		!mu < 0.25/timestep
		mu = 0.9*0.25/maxstep

!-----------  Mask salt area in tomograd  ------------------------------------------
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
		if (exist_file('tomograd'  )) call sep_write(tomograd_ns,  'tomograd' )
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
		call salt_masker(tomograd_ns)
		call smoother(tomograd_ns, tomograd, smooth_rad)

!-----------  Calculate bounddata  -------------------------------------------------
		call phi_grad_operator(phi_grad, bounddata, junk, s, scaling, .false.)	! Forward: phigrad  ==> bounddata

!-----------  Calculate tomodata  --------------------------------------------------
		call L_lop2(.false., .false., tomograd, tomodata)			! Forward: tomograd ==> tomodata

!-----------  Find alpha & beta in residual space  ---------------------------------
		phi_update = phi_grad + mu*drlse_term

		call Lsearch_init( tomodata, bounddata, residual )
		call L_planesearch( alpha, beta )
		call NLsearch_init( vel_back_ns, phi_in, phi_update, tomograd, maxstep, alpha, beta )
		call NL_linesearch( gamma )

write(0,*) " alpha b4 = ", alpha, "beta b4 = ", beta, "gamma = ", gamma
		alpha = gamma*alpha
		beta  = gamma*beta

!-----------  Scale down values if greater than CFL condition  ---------------------------------
		if ( beta > maxstep ) then
			alpha = alpha*(beta/maxstep)
			beta = 0.97*maxstep
		end if		

		alphaarray(1)= alpha
		betaarray(1) = beta
write(0,*) " alpha = ", alpha, "beta = ", beta, "maxstep =", maxstep


!-----------  Update the velocity model  -------------------------------------------
		vel_out = vel_back_ns + alpha*(tomograd)

!-----------  Update the implicit surface  -----------------------------------------
		phi_out  = phi_in + beta*(phi_update)

!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
		if (exist_file('alpha'     )) call sep_write( alphaarray , 'alpha' )
		if (exist_file('beta'     )) call sep_write( betaarray , 'beta' )
		if (exist_file('objfuncval' )) call sep_write( objarray , 'objfuncval' )

		if (exist_file('phi_out' )) call sep_write(phi_out,'phi_out' )
		if (exist_file('residual'  )) call sep_write(residual,  'residual' )
		if (exist_file('vel_full'  )) call sep_write(vel_full, 'vel_full')
		if (exist_file('vel_back'  )) call sep_write(vel_back_ns,'vel_back')
		if (exist_file('s'         )) call sep_write(s,  's'               )
		if (exist_file('drlse_term')) call sep_write(drlse_term, 'drlse_term')
		if (exist_file('phi_grad'  )) call sep_write(phi_grad,  'phi_grad' )
		if (exist_file('tomodata'  )) call sep_write(tomodata,  'tomodata' )
		if (exist_file('bounddata' )) call sep_write(bounddata, 'bounddata')
!@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@
	end subroutine




!/////////////////////////////////////////////////////////////////////////////////////
	subroutine velmodel_clip( velmodel, minvel, maxvel )
		real				:: minvel, maxvel
		real, dimension(:,:), target	:: velmodel		
		do j=1,nx
			do i=1,nz
				if (velmodel(i,j) > maxvel) then
					velmodel(i,j)=maxvel
				else if (velmodel(i,j) < minvel) then
					velmodel(i,j)=minvel
				end if
			end do
		end do
	end subroutine



!/////////////////////////////////////////////////////////////////////////////////////
	subroutine salt_masker(tomograd)
		real, dimension(:,:), target	:: tomograd
		do i=1,nz
			do j=1,nx
				if (phi_in(i,j) > 0.0 ) then
					tomograd(i,j) = 0.0
				end if
			end do
		end do
	end subroutine



end module

